from django.apps import (
    apps,
)
from django.db import (
    models,
)


def make_fk(model, field_name, **kwargs):
    field = model._meta.get_field(field_name)
    _, db_column = field.get_attname_column()

    unique = kwargs.pop('unique', False)
    kwargs.setdefault('on_delete', models.DO_NOTHING)
    kwargs.setdefault('related_name', '+')
    kwargs.setdefault('verbose_name', field.verbose_name)
    kwargs.setdefault('null', field.null)
    kwargs.setdefault('blank', field.blank)
    kwargs.setdefault('db_column', db_column)

    if unique:
        new_field_type = models.OneToOneField
    else:
        new_field_type = models.ForeignKey

    new_field = new_field_type(**kwargs)

    model._meta.local_fields.remove(field)
    new_field.contribute_to_class(model, field_name)


def add_params(addr_model, params_model):

    @property
    def params(self):
        Param = apps.get_model(params_model)

        class ParamManager(Param._default_manager.__class__):
            def get_queryset(manager):
                return super().get_queryset().filter(objectid=self.objectid_id)

        manager = ParamManager()
        manager.model = Param

        return manager

    addr_model.params = params


class RegionCodeModelMixin(models.Model):
    region_code = models.SmallIntegerField(verbose_name='Код региона')

    class Meta:
        abstract = True



from m3_gar.base_models import (
    AddrObj as BaseAddrObj,
    AddrObjDivision as BaseAddrObjDivision,
    AddrObjTypes as BaseAddrObjTypes,
)
from m3_gar.models.op_type import (
    OperationTypes,
)
from m3_gar.models.reestr import (
    ReestrObjects,
)
from m3_gar.models.util import (
    RegionCodeModelMixin,
    add_params,
    make_fk,
)


__all__ = ['AddrObj', 'AddrObjDivision', 'AddrObjTypes']


class AddrObjDivision(BaseAddrObjDivision, RegionCodeModelMixin):
    """
    Сведения по операциям переподчинения
    """
    class Meta:
        verbose_name = 'Операция переподчинения'
        verbose_name_plural = 'Операции переподчинения'


class AddrObjTypes(BaseAddrObjTypes):
    """
    Сведения по типам адресных объектов
    """
    class Meta:
        verbose_name = 'Тип адресного объекта'
        verbose_name_plural = 'Типы адресных объектов'


class AddrObj(BaseAddrObj, RegionCodeModelMixin):
    """
    Сведения классификатора адресообразующих элементов
    """
    class Meta:
        verbose_name = 'Адресообразующий элемент'
        verbose_name_plural = 'Адресообразующие элементы'


# На момент описания моделей AddrObjTypes никак не связывается с AddrObj
# Существующее поле AddrObj.typename - текстовое представление (ул, пер, г, и т.п.)
# make_fk(AddrObj, '???', to=AddrObjTypes, null=True, blank=True)

make_fk(AddrObjDivision, 'parentid', to=ReestrObjects, db_constraint=False)
make_fk(AddrObjDivision, 'childid', to=ReestrObjects, db_constraint=False)

make_fk(AddrObj, 'opertypeid', to=OperationTypes)
make_fk(AddrObj, 'objectid', to=ReestrObjects)

add_params(AddrObj, 'm3_gar.AddrObjParams')



from django.db import (
    models,
)

from m3_gar.base_models import (
    AdmHierarchy as BaseAdmHierarchy,
    MunHierarchy as BaseMunHierarchy,
)
from m3_gar.models.reestr import (
    ReestrObjects,
)
from m3_gar.models.util import (
    RegionCodeModelMixin,
    make_fk,
)


__all__ = ['AdmHierarchy', 'MunHierarchy']


class Hierarchy:
    """
    Базовый тип сведений по иерархии
    """

    @staticmethod
    def get_shortname_map():
        return {
            subclass.get_shortname(): subclass
            for subclass in Hierarchy.__subclasses__()
        }

    @classmethod
    def get_shortname(cls):
        shortname = ''

        if cls.__name__.endswith(Hierarchy.__name__):
            shortname = cls.__name__[:-len(Hierarchy.__name__)]

        return shortname.lower()


class AdmHierarchy(BaseAdmHierarchy, Hierarchy, RegionCodeModelMixin):
    """
    Сведения по иерархии в административном делении
    """
    class Meta:
        verbose_name = verbose_name_plural = 'Иерархия в административном делении'


class MunHierarchy(BaseMunHierarchy, Hierarchy, RegionCodeModelMixin):
    """
    Сведения по иерархии в муниципальном делении
    """
    class Meta:
        verbose_name = verbose_name_plural = 'Иерархия в муниципальном делении'


make_fk(AdmHierarchy, 'objectid', to=ReestrObjects)
make_fk(MunHierarchy, 'objectid', to=ReestrObjects)

make_fk(AdmHierarchy, 'parentobjid', to=ReestrObjects)
make_fk(MunHierarchy, 'parentobjid', to=ReestrObjects)



from copy import (
    deepcopy,
)

from django.db import (
    connections,
    models,
)

from m3_gar import (
    config,
)


def get_simple_field(field):
    """
    Возвращает упрощённую копию переданного поля с отключенными индексами,
    уникальными ограничениями и проверками внешних ключей
    """

    new_field = deepcopy(field)

    new_field._unique = False
    new_field.db_index = False

    if any((
        isinstance(field, models.ForeignKey),
        isinstance(field, models.ManyToManyField) and field.remote_field.through is None,
    )):
        new_field.db_constraint = False

    return new_field


def get_simple_fields(model):
    """
    Возвращает пары (поле, упрощённая копия поля) для всех собственных полей модели
    """

    fields = model._meta.get_fields(include_parents=False, include_hidden=False)

    for field in fields:
        if field.concrete:
            yield field, get_simple_field(field)


def change_constraints_for_model(model, field_from, field_to):
    """
    Производит преобразование поля field_from на field_to модели model
    в БД для записей ГАР

    Args:
        model: модель
        field_from: исходное поле
        field_to: новое поле

    Returns:
        список выполненных sql-запросов

    """
    con = connections[config.DATABASE_ALIAS]
    ed = con.schema_editor(collect_sql=True)

    ed.alter_field(model, field_from, field_to)

    return ed.collected_sql


def remove_constraints_from_model(model):
    """
    Удаляет ограничения и индексы модели
    """
    for field, simple_field in get_simple_fields(model=model):
        yield change_constraints_for_model(model=model, field_from=field, field_to=simple_field)


def restore_constraints_for_model(model):
    """
    Восстанавливает ограничения и индексы модели
    """
    for field, simple_field in get_simple_fields(model=model):
        yield change_constraints_for_model(model=model, field_from=simple_field, field_to=field)
